> module Powerlist where
> import Prelude hiding (succ, sum)
> import Rank2 (Functor2, Mu2, In, it, Nu2, Coit, out)

> pair :: (a -> b) -> (a,a) -> (b,b)
> pair f (a1, a2) = (f a1, f a2)

The constructor of pure kind 2 which we need to define powerlists.  We
take the freedom and use labelled sums.

> data PListF f a = Zero a | Succ (f (a, a))
> 
> instance Functor2 PListF where
>     ffmap s f (Zero a) = Zero (f a)
>     ffmap s f (Succ l) = Succ (s (pair f) l)

Inductive type of powerlists.

> type PList a = Mu2 PListF a

Powerlist constructors.

> zero :: a -> PList a
> zero a = In (Zero a)
>
> succ :: PList (a,a) -> PList a
> succ l = In (Succ l)


Summing up a powerlist of Integers.

We make use of the right Kan extension.
Unfortunately, we cannot use a type definition and need a datatype instead.

> newtype RKanInt a = RKanInt ((a -> Integer) -> Integer)

Step term.

> stepSum :: PListF RKanInt a -> RKanInt a
> stepSum (Zero a) = RKanInt (\f ->  f a)
> stepSum (Succ (RKanInt l)) = RKanInt (\f -> l (\ (a1,a2) -> f a1 + f a2))

> sum' :: (a -> b) -> PList a -> RKanInt b
> sum' = it stepSum

> sum :: PList Integer -> Integer
>
> sum l = k id 
>     where (RKanInt k) = sum' id l

\comm{

Convert to list (simple version).

> unzipMap :: (a -> b) -> [(a,a)] -> [b]
> unzipMap f [] = []
> unzipMap f ((a1,a2):as) = f a1 : f a2 : unzipMap f as

> stepListS :: PListF [ ] a -> [a]
> stepListS (Zero a) = [a]
> stepListS (Succ l) = unzipMap id l

> toListS :: PList a -> [a]
> toListS l = it stepListS id l 

Convert to list.

> newtype RKanList b = RKanList (forall c. (b -> [c]) -> [c])
>
> stepList :: PListF RKanList a -> RKanList a
>
> stepList (Zero a) = RKanList (\f -> f a)
> stepList (Succ (RKanList l)) = RKanList (\f -> 
>   l (\ (a1,a2) -> f a1 ++ f a2))
>
> toList :: PList a -> [a]
> toList l = k (\a -> [a]) where (RKanList k) = it stepList id l

Tests.

> mkPairs :: [a] -> [(a,a)]
> mkPairs [] = []
> mkPairs (a1:a2:as) = (a1,a2) : mkPairs as

> mkPList :: [a] -> PList a
> mkPList []  = error "not a powerlist"
> mkPList [a] = zero a
> mkPList l   = succ (mkPList (mkPairs l))

> l1 = mkPList [1,2,3,4]
> l2 = mkPList [0,1,2,3,4,5,6,7]

> l1sum = sum l1
> l2sum = sum l2

> pl1  = succ (succ (succ (zero (((1, 2), (3, 4)), ((5, 6), (7,8))))))
> tl1  = toList pl1
> tl1S = toListS pl1
 
end of comment } 